import re
from pathlib import Path

import click
import yaml

from kks.util.common import find_workspace, find_problem_rootdir, config_directory


target_file = 'targets.yaml'

targets_version = 4

global_comment = '# This is the default config file, it is used in any subdirectory of the workspace.\n'\
                 '# You can modify the default target (the changes will be applied only in this workspace).\n'\
                 '# Also, you can add/overwrite targets by creating another "targets.yaml" file in your working directory.\n'\
                 '\n'

class Target:
    def __init__(self, name, settings):
        self.name = name
        self.files = settings.get('files')
        self.compiler = settings.get('compiler')
        self.flags = settings.get('flags')
        self.libs = settings.get('libs')
        self.asm64bit = settings.get('asm64bit')
        self.default_asan = settings.get('default_asan')
        self.out = settings.get('out')

        # option is not set or the list contains "DEFAULT" as first item
        # if the list is empty, we shouldn't replace it (e.g. if this target excludes all libs)
        self.need_default = any(arr is None or (arr and arr[0] == 'DEFAULT') for arr in [self.files, self.flags, self.libs])\
                            or self.compiler is None or self.out is None or self.asm64bit is None or self.default_asan is None
        # custom default target may still have some fields not set, so we will have to get them from a higher-level config (workspace root or package-provided)

    def __str__(self):
        return f'Target("{self.name}", compiler="{self.compiler}", flags={self.flags}, files={self.files}, libs={self.libs}, asm64bit={self.asm64bit}, default_asan={self.default_asan}, out="{self.out}")'

    def replace_macros_add_missing(self, problem, default_target):
        def modify(x):
            return x.replace('TASKNAME', problem)

        def modify_list(lst, default):
            if lst is None:
                return default
            if len(lst) == 0:
                return lst
            if lst[0] == 'DEFAULT':
                return [modify(e) for e in default + lst[1:]]
            return [modify(e) for e in lst]

        self.compiler = self.compiler or default_target.compiler
        self.files = modify_list(self.files, default_target.files)
        self.flags = modify_list(self.flags, default_target.flags)
        self.libs = modify_list(self.libs, default_target.libs)
        self.out = modify(self.out) if self.out is not None else default_target.out  # "or" doesn't work
        self.asm64bit = self.asm64bit if self.asm64bit is not None else default_target.asm64bit
        self.default_asan = self.default_asan if self.default_asan is not None else default_target.default_asan


def _copy_default(dest):
    from pkg_resources import resource_stream
    default_cfg = resource_stream('kks', f'data/{target_file}').read()
    dest.write_bytes(default_cfg.replace(b'KKS_TARGETS_VERSION', str(targets_version).encode(), 1))


cfg_file = config_directory() / target_file
if not cfg_file.exists():
    _copy_default(cfg_file)
package_cfg = yaml.safe_load(cfg_file.read_text())
if package_cfg['__version__'] != targets_version:  # update
    _copy_default(cfg_file)
    package_cfg = yaml.safe_load(cfg_file.read_text())

# package_default should have all fields set
package_default = Target('default', package_cfg['default'])


def check_version(cfg_file, cfg, is_global=False):
    if cfg.get('__version__', package_cfg['__version__']) == package_cfg['__version__']:
        return

    old_version = cfg['__version__']
    new_version = package_cfg['__version__']
    click.echo(click.style(str(cfg_file.absolute()), fg='blue', bold=True) +
               click.style(' is outdated. You can run "kks init --config=update" if you want to update the default target manually (see README on Github)', fg='yellow'))
    cfg_file.write_text(re.sub(rf'^(__version__: +){old_version}', rf'\g<1>{new_version}', cfg_file.read_text(), 1, re.M))


def find_target(name):
    def load_config(config_file):
        with config_file.open('r') as f:
            try:
                config = yaml.safe_load(f)
                return config
            except yaml.parser.ParserError:
                click.secho(f'Cannot parse {config_file.absolute()}', fg='red', err=True)
                return None

    def get_target(config, target_name):
        if config is None:
            return None
        if target_name in config:
            return Target(target_name, config[target_name])
        return None

    cwd = Path.cwd()
    workspace = find_workspace()
    rootdir = find_problem_rootdir()
    problem = '{}-{}'.format(*rootdir.parts[-2:]) if rootdir is not None else 'sm00-0'

    if workspace is not None:
        root_file = workspace / target_file  # we load root config even if we won't need it
        root_cfg = load_config(root_file) if root_file.is_file() else None
        if root_cfg is not None:
            check_version(root_file, root_cfg, True)
    else:
        root_cfg = None

    if workspace is None or cwd != workspace:
        local_file = cwd / target_file
        local_cfg = load_config(local_file) if local_file.is_file() else None
        if local_cfg is not None:
            check_version(local_file, local_cfg)
    else:
        local_cfg = None

    local_target = get_target(local_cfg, name)
    if local_target is not None:
        default = package_default
        if local_target.need_default:
            default = get_target(local_cfg, 'default') or get_target(root_cfg, 'default') or package_default
            if default.need_default:  # custom default target may have some missing fields
                default.replace_macros_add_missing(problem, package_default)
        local_target.replace_macros_add_missing(problem, default)
        return local_target

    root_target = get_target(root_cfg, name)
    if root_target is not None:
        default = package_default
        if root_target.need_default:
            # it makes no sense to look for default in CWD if the target is in workspace root
            default = get_target(root_cfg, 'default') or package_default
            # TODO optimize? if root_target is "default", then default == root_target and the next condition is always true
            if default.need_default:
                default.replace_macros_add_missing(problem, package_default)
        root_target.replace_macros_add_missing(problem, default)
        return root_target

    if name == 'default':
        return package_default

    package_target = get_target(package_cfg, name)
    if package_target is None:
        # not found
        return None

    package_target.replace_macros_add_missing(problem, package_default)
    return package_target



import click

from kks.ejudge import ejudge_summary
from kks.util.ejudge import EjudgeSession
from kks.errors import AuthError


@click.command(short_help='Parse and display task status')
@click.argument('filters', nargs=-1)
def status(filters):
    """
    Parse and display task status

    If any FILTERS are specified, show status only for tasks with matching prefixes/names
    """

    try:
        session = EjudgeSession()
    except AuthError:
        return

    problems = ejudge_summary(session)
    if problems is None:
        return

    if filters:
        problems = [p for p in problems if any(p.short_name.startswith(f) for f in filters)]
        if not problems:
            click.secho('Nothing found')
            return

    row_format = "{:8} {:35} {:20} {:>5}"

    click.secho(row_format.format("Alias", "Name", "Status", "Score"), bold=True)

    for problem in problems:
        string = row_format\
            .format(problem.short_name, problem.name, problem.status, problem.score or '')

        click.secho(string, fg=problem.color(), bold=problem.bold())

    click.secho()



import re
from collections import namedtuple
from itertools import groupby

import click
from tqdm import tqdm

from kks.ejudge import LinkTypes, Status, ejudge_standings, ejudge_summary, get_problem_info, extract_contest_name
from kks.util.common import read_config, write_config, set_boolean_option, get_boolean_option, has_boolean_option
from kks.util.ejudge import EjudgeSession
from kks.errors import AuthError
from kks.util.cache import Cache
from kks.util.stat import send_standings, get_global_standings

GLOBAL_OPT_OUT = 'global-opt-out'

Problem = namedtuple('Problem', ['href', 'short_name', 'contest'])  # used for caching the problem list


ROW_FORMAT = "{:>6}  {:25} {:>7} {:>6}{}"
PREFIX_LENGTH = sum([6, 2, 25, 1, 7, 1, 6])
CONTEST_DELIMITER = ' | '


@click.command(short_help='Parse and display user standings')
@click.option('-l', '--last', type=int,
              help='Print result of last N contest')
@click.option('-a', '--all', 'all_', is_flag=True,
              help='Print result of all contests')
@click.option('-c', '--contest', 'contests', type=str, multiple=True,
              help='Print the results of the selected contest')
@click.option('-m', '--max',  'max_', is_flag=True,
              help='Print maximal possible scores (based on current deadlines)')
@click.option('-nc', '--no-cache', is_flag=True,
              help='Clear cache and reload task info (used with --max)')
@click.option('-g', '--global', 'global_', is_flag=True,
              help='Use global standings instead of group one. May be outdated')
@click.option('--global-opt-out', is_flag=True,
              help='Opt out from submitting your group results')
def top(last, contests, all_, max_, no_cache, global_, global_opt_out):
    """
    Parse and display user standings

    \b
    Example usage:
        kks top
        kks top --all
        kks top -c sm01 -c sm02
        kks top --last 2
    """

    config = read_config()

    if global_opt_out:
        if click.confirm('Do you really want to opt out from sending your group standings to kks.darkkeks.me?'):
            set_boolean_option(config, GLOBAL_OPT_OUT, True)
            write_config(config)
            click.secho('Successfully disabled standings sending. You can always enable sending by manually editing '
                        '~/.kks/config.ini', color='red', err=True)
        return
    elif not has_boolean_option(config, GLOBAL_OPT_OUT):
        click.secho('Standings can be sent for aggregation to kks api. '
                    'This allows us to create global standings (you can try it out with kks top --global)', err=True)
        click.secho('You can always disable sending using kks top --global-opt-out', err=True)
        if click.confirm('Do you want to send group standings to kks api?', default=True):
            click.secho('Thanks a lot for you contribution! We appreciate it!', color='green', err=True)
            set_boolean_option(config, GLOBAL_OPT_OUT, False)
            write_config(config)
        else:
            click.secho('Successfully disabled standings sending. You can always enable sending by manually editing '
                        '~/.kks/config.ini', color='red', err=True)
            set_boolean_option(config, GLOBAL_OPT_OUT, True)
            write_config(config)

    try:
        session = EjudgeSession()
    except AuthError:
        return

    standings = ejudge_standings(session)
    if standings is None:
        return

    if not get_boolean_option(config, GLOBAL_OPT_OUT):
        if not send_standings(standings):
            click.secho('Failed to send standings to kks api', color='yellow', err=True)

    if global_:
        standings = get_global_standings()
        if standings is None:
            click.secho('Standings are not available now :(', color='yellow', err=True)
            return

    if max_:
        standings = estimate_max(standings, session, no_cache)

    display_standings(standings, last, contests, all_)


def display_standings(standings, last, contests, all_):
    contests = select_contests(standings, last, contests, all_)
    if contests is None:
        return

    contest_widths = get_contest_widths(standings.contests, standings.tasks_by_contest)
    contests_header = ''.join([
        CONTEST_DELIMITER + contest.ljust(contest_widths[contest], ' ')
        for contest in contests
    ])
    header = ROW_FORMAT.format("Place", "User", "Solved", "Score", contests_header)

    click.secho(header, fg='white', bold=True)

    for row in standings.rows:
        tasks = ''.join([
            click.style(CONTEST_DELIMITER, fg=row.color(), bold=row.bold()) + ' '.join([
                click.style('{:>3}'.format(task.table_score() or ''), fg=task.color(), bold=task.bold())
                for task in tasks
            ])
            for contest, tasks in groupby(row.tasks, lambda task: task.contest)
            if contest in contests
        ])

        string = ROW_FORMAT.format(row.place, row.user, row.solved, row.score, tasks)
        click.secho(string, fg=row.color(), bold=row.bold())


def select_contests(standings, last, contests, all_):
    has_last = last is not None
    has_contests = len(contests) > 0

    if sum([has_last, has_contests, all_]) > 1:
        click.secho('Arguments are exclusive, specify no more than one')
        return None

    if all_:
        return standings.contests

    if has_contests:
        def is_available(contest):
            if contest not in standings.contests:
                click.echo(
                        click.style('Contest ', fg='red') +
                        click.style(contest, fg='blue', bold=True) +
                        click.style(' not found!', fg='red'),
                        err=True
                )
            return contest in standings.contests

        return list(filter(is_available, contests))

    last = last or get_default_contest_count(standings.contests, standings.tasks_by_contest)

    return standings.contests[-last:] if last > 0 else []


def get_default_contest_count(contests, tasks_by_contest):
    """По ширине терминала пытается определить, сколько колонок можно вывести"""
    (width, _) = click.get_terminal_size()

    delimiter_width = len(CONTEST_DELIMITER)
    contest_widths = get_contest_widths(contests, tasks_by_contest)

    width_sum = 0
    for i, contest in enumerate(contests[::-1]):
        width_sum += delimiter_width + contest_widths[contest]
        if width_sum + PREFIX_LENGTH > width:
            return i

    return len(contests)


def get_contest_widths(contests, tasks_by_contest):
    return {
        contest: len('100') * len(tasks_by_contest[contest]) + len(tasks_by_contest[contest]) - 1
        for contest in contests
    }


def estimate_max(standings, session, force_reload):
    # NOTE may produce incorrect results for "krxx" contests (they may be reopened?)

    sid_regex = re.compile('/S([0-9a-f]{16})')
    sid = sid_regex.search(session.links.get(LinkTypes.USER_STANDINGS)).group(1)

    def cached_problem(problem):
        href = re.sub(sid_regex, '/S__SID__', problem.href, count=1)
        return Problem(href, problem.short_name, extract_contest_name(problem.short_name))

    def with_fixed_href(problem):
        href = problem.href.replace('__SID__', sid)
        return Problem(href, problem.short_name, problem.contest)

    standings.rows = list(standings.rows)

    with Cache('problem_info', compress=True).load() as cache:
        if force_reload:
            cache.clear()

        problem_list = cache.get('problem_links', [])  # we can avoid loading summary

        with tqdm(total=len(standings.tasks), leave=False) as pbar:
            def with_progress(func, *args, **kwargs):
                result = func(*args, **kwargs)
                pbar.update(1)
                return result

            if len(problem_list) != len(standings.tasks) or \
                    any(problem.short_name != task.name for problem, task in zip(problem_list, standings.tasks)):
                problem_list = ejudge_summary(session)
                problem_list = [cached_problem(p) for p in problem_list]
                cache.set('problem_links', problem_list)
            problems = [with_progress(get_problem_info, with_fixed_href(problem), cache, session) for problem in problem_list]

    for row in standings.rows:
        for task_score, problem in zip(row.tasks, problems):
            if task_score.score is None or task_score.score == '0':
                if task_score.status == Status.REJECTED:
                    max_score = problem.full_score
                else:
                    max_score = problem.full_score - problem.current_penalty
                    if task_score.score == '0':
                        max_score -= problem.run_penalty
                        # actually may be lower
                if max_score > 0:
                    row.solved += 1
                    row.score += max_score
                    task_score.score = max_score
                    task_score.status = Status.REVIEW

    standings.rows.sort(key=lambda x: (x.score, x.solved), reverse=True)
    for i, row in enumerate(standings.rows):
        row.place = i + 1

    return standings



import json
import pickle
import re
import click

from kks import __version__
from kks.ejudge import LinkTypes, AuthData, get_contest_url
from kks.util.common import config_directory, read_config, write_config
from kks.errors import AuthError, APIError


def load_auth_data():
    config = read_config()
    if not config.has_section('Auth'):
        return None
    auth = config['Auth']
    if 'login' in auth and 'contest' in auth:
        return AuthData(auth['login'], auth['contest'], auth.get('password', None))
    return None


def save_auth_data(auth_data, store_password=True):
    config = read_config()
    config['Auth'] = {
        'login': auth_data.login,
        'contest': auth_data.contest_id
    }

    if store_password and auth_data.password is not None:
        config['Auth']['password'] = auth_data.password

    write_config(config)


def load_links():
    config = read_config()
    if config.has_section('Links'):
        return config['Links']
    return None


def save_links(links):
    config = read_config()
    config['Links'] = links
    write_config(config)


def load_session():
    import requests

    cookies = config_directory() / 'cookies.pickle'
    if not cookies.is_file():
        return None

    session = requests.session()
    with open(cookies, 'rb') as f:
        try:
            cookies = pickle.load(f)
        except Exception:
            return None
        session.cookies.update(cookies)
    return session


def store_session(session):
    cookies = config_directory() / 'cookies.pickle'
    with open(cookies, 'wb') as f:
        pickle.dump(session.cookies, f)


class RunStatus:
    """A (very) limited wrapper class for responses from "run-status-json" method"""

    COMPILING = 98  # from github.com/blackav/ejudge-fuse
    COMPILED = 97
    RUNNING = 96

    # this group is also used in test results
    OK = 0
    CE = 1
    RE = 2
    TL = 3
    PE = 4
    WA = 5
    ML = 12

    CHECK_FAILED = 6
    PARTIAL = 7
    ACCEPTED = 8
    IGNORED = 9
    PENDING = 11

    PENDING_REVIEW = 16
    REJECTED = 17
    SKIPPED = 18

    # There are more, but only these were seen on caos.ejudge.ru

    _descriptions = {
        COMPILING: 'Compiling',
        COMPILED: 'Compiled',
        RUNNING: 'Running',
        OK: 'OK',
        CE: 'Compilation error',
        RE: 'Runtime error',
        TL: 'Time limit exceeded',
        PE: 'Presentation error',
        WA: 'Wrong answer',
        ML: 'Memory limit exceeded',
        CHECK_FAILED: 'Check failed',
        PARTIAL: 'Partial solution',
        ACCEPTED: 'Accepted for testing',
        IGNORED: 'Ignored',
        PENDING: 'Pending check',
        PENDING_REVIEW: 'Pending review',
        REJECTED: 'Rejected',
        SKIPPED: 'Skipped'
    }

    @staticmethod
    def get_description(status_code):
        return RunStatus._descriptions.get(status_code, f'Unknown status {status_code}')

    def __init__(self, run_status):
        self.status = run_status['run']['status']
        self.tests = []
        if 'testing_report' in run_status and 'tests' in run_status['testing_report']:
            self.tests = run_status['testing_report']['tests']

    def is_testing(self):
        return self.status >= 95 and self.status <= 99

    def __str__(self):
        return self.get_description(self.status)

    def with_tests(self, failed_only=False):
        if not self.tests:
            return str(self)

        def test_descr(test):
            return f"{test['num']} - {self.get_description(test['status'])}"

        if failed_only:
            test_results = '\n'.join(map(test_descr, [test for test in self.tests if test['status'] not in [self.OK, self.SKIPPED]]))
        else:
            test_results = '\n'.join(map(test_descr, self.tests))
        return f'{self}\n{test_results}'


class API:
    def __init__(self, sids=None):
        import requests

        self._prefix = 'https://caos.ejudge.ru/cgi-bin/'
        self._http = requests.Session()
        self._http.headers = {'User-Agent': f'kokos/{__version__}'}

        self._sids = sids

    def _request(self, url, need_json, **kwargs):
        resp = self._http.post(url, **kwargs)  # all methods accept POST requests
        resp.encoding = 'utf-8'  # ejudge doesn't set encoding header
        try:
            # all methods return errors in json
            data = json.loads(resp.content)
        except ValueError as e:
            if not need_json:
                return resp.content
            raise APIError(f'Invalid response. resp={resp.content}, err={e}', APIError.INVALID_RESPONSE)

        # if a submission is a valid JSON file, then api.download_run will fail
        if not need_json or not data['ok']:
            err = data.get('error', {})
            raise APIError(err.get('message', 'Unknown error'), err.get('num', APIError.UNKNOWN))
        return data['result']

    def _api_method(self, path, action, sids=None, need_json=True, **kwargs):
        """
        Allowed values for `sids`:
          - None (use self._sids)
          - {} (don't use sids)
          - {'EJSID': str, 'SID': str}
        """

        url = self._prefix + path

        data = kwargs.setdefault('data', {})
        data.update({'action': action, 'json': 1})
        if sids is None:
            sids = self._sids
        data.update(sids)

        return self._request(url, need_json, **kwargs)

    def auth(self, creds):
        """get new sids"""
        # NOTE is 1step auth possible?

        def extract_sids(resp):
            return {'EJSID': resp['EJSID'], 'SID': resp['SID']}

        sids = extract_sids(self.login(creds.login, creds.password))
        self._sids = extract_sids(self.enter_contest(sids, creds.contest_id))

    def login(self, login, password):
        """get sids for enter_contest method"""
        data = {
            'login': login,
            'password': password,
        }
        return self._api_method('register', 'login-json', {}, data=data)

    def enter_contest(self, sids, contest_id):
        data = {
            'contest_id': contest_id
        }
        return self._api_method('register', 'enter-contest-json', sids, data=data)

    def contest_status(self):
        return self._api_method('client', 'contest-status-json')

    def problem_status(self, prob_id):
        data = {
            'problem': int(prob_id)
        }
        return self._api_method('client', 'problem-status-json', data=data)

    def problem_statement(self, prob_id):
        data = {
            'problem': int(prob_id)
        }
        return self._api_method('client', 'problem-statement-json', data=data, need_json=False)

    def list_runs(self, prob_id=None):
        # newest runs go first
        # if no prob_id is passed then all runs are returned (useful for sync?)
        if prob_id is None:
            return self._api_method('client', 'list-runs-json')['runs']
        data = {
            'prob_id': int(prob_id)
        }
        return self._api_method('client', 'list-runs-json', data=data)['runs']

    def run_status(self, run_id):
        data = {
            'run_id': int(run_id)
        }
        return self._api_method('client', 'run-status-json', data=data)

    def download_run(self, run_id):
        data = {
            'run_id': int(run_id)
        }
        return self._api_method('client', 'download-run', data=data, need_json=False)

    def run_messages(self, run_id):
        data = {
            'run_id': int(run_id)
        }
        return self._api_method('client', 'run-messages-json', data=data)

    # run-test-json - test results? unknown params

    def submit(self, prob_id, file, lang):
        data = {
            'prob_id': int(prob_id),
        }
        if lang is not None:  # NOTE may possibly break on problems without lang (see sm01-3)
            data['lang_id'] = int(lang)

        files = {
            'file': (file.name, open(file, 'rb'))
        }
        return self._api_method('client', 'submit-run', data=data, files=files)


class EjudgeSession:
    sid_regex = re.compile('/S([0-9a-f]{16})')

    def __init__(self, auth=True):
        self._sids = {'SID': None, 'EJSID': None}

        self.http = load_session()
        if self.http is not None:
            self.links = load_links()

        if self.http is None or self.links is None:
            if auth:
                self.auth()
        else:
            self._update_sid()

    def auth(self, auth_data=None, store_password=None):
        if auth_data is None:
            click.secho('Cookies are either missing or invalid, trying to auth with saved data', fg='yellow', err=True)
            auth_data = load_auth_data()

            if auth_data is not None:
                store_password = True
                if auth_data.password is None:
                    store_password = False
                    auth_data.password = click.prompt('Password', hide_input=True)

        if auth_data is None:
            click.secho('No valid cookies or auth data, please use "kks auth" to log in', fg='yellow', err=True)
            raise AuthError()

        import requests
        from bs4 import BeautifulSoup

        if self.http is None:
            self.http = requests.session()
        else:
            self.http.cookies.clear()
        url = get_contest_url(auth_data)
        page = self.http.post(url, data={
            'login': auth_data.login,
            'password': auth_data.password
        })

        if page.status_code != requests.codes.ok:
            click.secho(f'Failed to authenticate (status code {page.status_code})', err=True, fg='red')
            raise AuthError()

        soup = BeautifulSoup(page.content, 'html.parser')

        if 'Invalid contest' in soup.text or 'invalid contest_id' in soup.text:
            click.secho(f'Invalid contest (contest id {auth_data.contest_id})', fg='red', err=True)
            raise AuthError()

        if 'Permission denied' in soup.text:
            click.secho('Permission denied (invalid username, password or contest id)', fg='red', err=True)
            raise AuthError()

        buttons = soup.find_all('a', {'class': 'menu'}, href=True)

        self.links = {
            button.text: button['href']
            for button in buttons
        }

        if self.links is None:
            click.secho('Auth data is invalid, use "kks auth" to authorize', fg='red', err=True)
            raise AuthError()

        save_auth_data(auth_data, store_password)
        save_links(self.links)
        store_session(self.http)
        self._update_sid()

    def api(self):
        """
        Create an API wrapper with (EJ)SID from this session
        If cookies are outdated, api requests will raise an APIError
        If api is used before any session requests are performed, use EjudgeSession.with_auth for the first request
        Example:
        >>> api = session.api()
        >>> problem = session.with_auth(api.problem_status, 123)
        >>> ...
        >>> info = api.contest_status()  # cookies are up to date
        """
        return API(self._sids)

    def with_auth(self, api_method, *args, **kwargs):
        """
        api_method should only a method of API object that was created with .api() method of this instance
        """
        try:
            return api_method(*args, **kwargs)
        except APIError as e:
            if e.code != APIError.INVALID_SESSION:
                raise e
            self.auth()
            return api_method(*args, **kwargs)

    def _update_sid(self):
        sid = EjudgeSession.sid_regex.search(self.links.get(LinkTypes.SUMMARY)).group(1)
        self._sids.update({'SID': sid, 'EJSID':  self.http.cookies['EJSID']})

    def modify_url(self, url):
        return re.sub(EjudgeSession.sid_regex, f"/S{self._sids['SID']}", url, count=1)

    def _request(self, method, url, *args, **kwargs):
        response = method(url, *args, **kwargs)
        if 'Invalid session' in response.text:
            self.auth()
            url = self.modify_url(url)
            response = method(url, *args, **kwargs)
        return response

    def get(self, url, *args, **kwargs):
        return self._request(self.http.get, url, *args, **kwargs)

    def post(self, url, *args, **kwargs):
        return self._request(self.http.post, url, *args, **kwargs)
